#include "keywords.h"

void init_lexer(const char *source) {
  lexer.start = source;
  lexer.current = source;
  lexer.line = 1;
}

int is_at_end() {
  return *lexer.current == '\0';
}

char advance() {
  return *lexer.current++;
}

char peek() {
  return *lexer.current;
}

char peek_next() {
  if (is_at_end()) return '\0';
  return lexer.current[1];
}

int match(char expected) {
  if (*lexer.current != expected) return 0;
  lexer.current++;
  return 1;
}

void skip_whitespace() {
  for (;;) {
    char c = peek();
    switch (c) {
      case ' ':
      case '\r':
      case '\t':
        advance();
        break;
      case '\n':
        lexer.line++;
        advance();
        break;
      case '!':
        if (peek_next() == '!') {
          while (peek() != '\n' && !is_at_end()) advance();
        } else {
          while (peek() != '\n' && !is_at_end()) advance();
        }
        break;
      default:
        return;
    }
  }
}

Token make_token(TokenType type) {
  Token token;
  token.type = type;
  token.start = lexer.start;
  token.length = (int)(lexer.current - lexer.start);
  token.line = lexer.line;
  return token;
}

Token error_token(const char *message) {
  Token token;
  token.type = TOKEN_ERROR;
  token.start = message;
  token.length = (int)strlen(message);
  token.line = lexer.line;
  return token;
}

int is_alpha(char c) {
  return isalpha(c) || c == '_';
}

int is_digit(char c) {
  return isdigit(c);
}

Token number() {
  while (is_digit(peek())) advance();

  if (peek() == '.' && is_digit(peek_next())) {
    advance();
    while (is_digit(peek())) advance();
    return make_token(TOKEN_FLOAT_LITERAL);
  }

  return make_token(TOKEN_INT_LITERAL);
}

Token string() {
  while (peek() != '"' && !is_at_end()) {
    if (peek() == '\n') lexer.line++;
    advance();
  }

  if (is_at_end()) return error_token("Unterminated string.");

  advance(); // Consume closing quote
  return make_token(TOKEN_STRING_LITERAL);
}

Token next_token() {
  skip_whitespace();
  lexer.start = lexer.current;

  if (is_at_end()) return make_token(TOKEN_EOF);

  char c = advance();

  if (is_alpha(c)) return identifier();
  if (is_digit(c)) return number();

  switch (c) {
    case '(': return make_token(TOKEN_LPAREN);
    case ')': return make_token(TOKEN_RPAREN);
    case '{': return make_token(TOKEN_LBRACE);
    case '}': return make_token(TOKEN_RBRACE);
    case '[': return make_token(TOKEN_LBRACKET);
    case ']': return make_token(TOKEN_RBRACKET);
    case ',': return make_token(TOKEN_COMMA);
    case '.': return make_token(TOKEN_DOT);
    case ':': return make_token(TOKEN_COLON);
    case ';': return make_token(TOKEN_SEMICOLON);
    case '+': return make_token(TOKEN_PLUS);
    case '-': return make_token(TOKEN_MINUS);
    case '*': return make_token(TOKEN_STAR);
    case '/': return make_token(TOKEN_SLASH);
    case '=': return make_token(TOKEN_EQ);
    case '"': return string();
    case '!':
      if (match('!')) return make_token(TOKEN_DOUBLE_BANG);
      return make_token(TOKEN_BANG);
  }

  return error_token("Unexpected character.");
}

const char *token_type_name(TokenType type) {
  switch (type) {
    case TOKEN_IDENTIFIER: return "identifier";
    case TOKEN_INT_LITERAL: return "int";
    case TOKEN_FLOAT_LITERAL: return "float";
    case TOKEN_STRING_LITERAL: return "string";
    case TOKEN_KEYWORD_TYPE: return "type";
    case TOKEN_KEYWORD_FN: return "fn";
    case TOKEN_KEYWORD_LET: return "let";
    case TOKEN_KEYWORD_CONST: return "const";
    case TOKEN_KEYWORD_IF: return "if";
    case TOKEN_KEYWORD_ELSE: return "else";
    case TOKEN_KEYWORD_WHILE: return "while";
    case TOKEN_KEYWORD_FOR: return "for";
    case TOKEN_KEYWORD_RETURN: return "return";
    case TOKEN_KEYWORD_USE: return "use";
    case TOKEN_OPERATOR_IS: return "is";
    case TOKEN_BANG: return "!";
    case TOKEN_DOUBLE_BANG: return "!!";
    case TOKEN_EQ: return "=";
    case TOKEN_DOT: return ".";
    case TOKEN_COMMA: return ",";
    case TOKEN_COLON: return ":";
    case TOKEN_SEMICOLON: return ";";
    case TOKEN_PLUS: return "+";
    case TOKEN_MINUS: return "-";
    case TOKEN_STAR: return "*";
    case TOKEN_SLASH: return "/";
    case TOKEN_LPAREN: return "(";
    case TOKEN_RPAREN: return ")";
    case TOKEN_LBRACE: return "{";
    case TOKEN_RBRACE: return "}";
    case TOKEN_LBRACKET: return "[";
    case TOKEN_RBRACKET: return "]";
    case TOKEN_EOF: return "eof";
    case TOKEN_ERROR: return "error";
    default: return "unknown";
  }
}


